/* Adapted from a "Style for an "Interest Group Note" */ /*
   Copyright 1997-2003 W3C (MIT, ERCIM, Keio). All Rights Reserved.
   The following software licensing rules apply:
   http://www.w3.org/Consortium/Legal/copyright-software */

body {
	padding: 2em 1em 2em 70px;
	margin: 0;
	font-family: sans-serif;
	color: black;
	background: white;
	background-position: top left;
	background-attachment: fixed;
	background-repeat: no-repeat;
}

blockquote { margin-top: 0pt; margin-bottom: 0pt; text-align: justify }

pre { border: 1.00pt solid #000000; padding: 0.02in; width:85%; }
    
blockquote pre { border: none; width: 100% }
    
p {text-indent: 15pt}
    
.footnotes { font-size: smaller; background-color:#fef; width: 70% }
    
address {text-align: right; text-indent: 15pt}

:link {
	color: #856347;
	background: transparent
}

table {
	border: solid;
	width: 100%;
	border-collapse: collapse;
}

td {
	border: 1px solid black;
}

TD.rowheader {
	width: 20%;	
}

/* Deprecated class/properties */
table.deprecated {
        border: dotted;
}

td.deprecated {
        color: #f88324;
}

h3.deprecated {
        color: #f88324;
}

:visited {
	color: #609;
	background: transparent
}

a:active {
	color: #C00;
	background: transparent
}

a:link img,a:visited img {
	border-style: none
}

/* no border on img links */
a img {
	color: white;
	vertical-align: middle;
	margin: 20;
}

/* trick to hide the border in Netscape 4 */
@media all { 
	/* hide the next rule from Netscape 4 */
	a img {
		color: inherit;
	} 
	/* undo the color change above */
}

th,td { 
	/* ns 4 */
	font-family: sans-serif;
}

h1,h2,h3,h4,h5,h6 {
	text-align: left
}

/* background should be transparent, but WebTV has a bug */
h1,h2,h3 {
	color: #CE5C00;
	background: white
}

h1 {
	font: 170% sans-serif
}

h2 {
	font: 140% sans-serif
}

h3 {
	font: 120% sans-serif
}

h4 {
	font: bold 100% sans-serif
}

h5 {
	font: italic 100% sans-serif
}

h6 {
	font: small-caps 100% sans-serif
}

.hide {
	display: none
}

div.head {
	margin-bottom: 1em
}

div.head h1 {
	margin-top: 2em;
	clear: both
}

div.head table {
	margin-left: 2em;
	margin-top: 2em
}

p.copyright {
	font-size: small
}

p.copyright small {
	font-size: small
}

@media screen { /* hide from IE3 */
	a [href]:hover {
		background: #ffa
	}
}

pre {
	margin-left: 2em
}

/*
p {
  margin-top: 0.6em;
  margin-bottom: 0.6em;
}
*/
dt,dd {
	margin-top: 0;
	margin-bottom: 0
}  /* opera 3.50 */
dt {
	font-weight: bold
}

pre,code {
	font-family: monospace
}  /* navigator 4 requires this */
ul.toc,ol.toc {
	list-style: disc; /* Mac NS has problem with 'none' */
	list-style: none;
}

@media aural {
	h1,h2,h3 {
		stress: 20;
		richness: 90
	}
	.hide {
		speak: none
	}
	p.copyright {
		volume: x-soft;
		speech-rate: x-fast
	}
	dt {
		pause-before: 20%
	}
	pre {
		speak-punctuation: code
	}
}

/* Copyright 1997-2003 W3C (MIT, ERCIM, Keio). All Rights Reserved.
   The following software licensing rules apply:
   http://www.w3.org/Consortium/Legal/copyright-software */
body {
	background-image: url(logo-maemo-ontology.png);
}